from silence.db.dal import query

from silence.logging.default_logger import logger

from silence.settings import settings

# Caches the columns of a table, to avoid repetitive queries
global TABLE_COLUMNS
TABLE_COLUMNS = {}

def get_tables():
    res = query(q = "SHOW FULL TABLES WHERE table_type = 'BASE TABLE';")
    tables = {}
    for table_data in res:
        # Grab the value that is not "BASE TABLE", which will be the name of the table
        table_name = next(x for x in table_data.values() if x != "BASE TABLE")
        tables[table_name] = get_table_cols(table_name)

    logger.debug("Tables in database: %s", str(tables))
    return tables

def get_views():
    res = query(q = "SHOW FULL TABLES WHERE table_type = 'VIEW';")
    views = {}
    for view_data in res:
        # Grab the value that is not "VIEW", which will be the name of the view
        view_name = next(x for x in view_data.values() if x != "VIEW")
        views[view_name] = get_table_cols(view_name)

    logger.debug("Views in database: %s", str(views))
    return views

def get_primary_key(table_name):
    t_pure = next(t for t in get_tables() if t.lower() == table_name.lower())
    primary = query(f"SHOW KEYS FROM {t_pure} WHERE Key_name = 'PRIMARY'")
    
    if primary:
        return primary[0]['Column_name']
    else:
        return None

def get_primary_key_views(view_name):
    all_primary_keys = []
    for table in get_tables():
        pk = get_primary_key(table)
        if pk:
            all_primary_keys.append(pk)
    
    primary_keys = []
    view_columns = query(f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{settings.DB_CONN['database']}' AND TABLE_NAME = '{view_name}'")
    for v in view_columns:
        if(v["COLUMN_NAME"] in all_primary_keys):
            primary_keys.append(v["COLUMN_NAME"])

    return primary_keys

def is_auto_increment(table_name, column_name):
    auto = query(f"SELECT EXTRA FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{settings.DB_CONN['database']}' AND TABLE_NAME = '{table_name}' AND COLUMN_NAME = '{column_name}'")
    res = auto[0]['EXTRA'] == 'auto_increment'
    return res

# Returns the list of names for the columns of a table, storing it
# after the first query for a given table
def get_table_cols(table_name):
    global TABLE_COLUMNS

    if table_name not in TABLE_COLUMNS:
        cols = query(f"SHOW COLUMNS FROM {table_name}")
        col_names = [col["Field"] for col in cols]
        TABLE_COLUMNS[table_name] = col_names
    
    return TABLE_COLUMNS[table_name]



from silence.settings import settings

from logging import StreamHandler
import threading
import platform

class MaybeBlockingHandler(StreamHandler):
    def __init__(self):
        super().__init__()

        # This is done to prevent colored log messages from overlapping
        # when running on Windows. This happens because stdout in Windows
        # must be flushed after every color change, which sometimes makes
        # colored log messages overlap.
        if settings.COLORED_OUTPUT and platform.system() == "Windows":
            self._write_lock = threading.Lock() 
        else:
            # Otherwise, we use a dummy "Lock" that does nothing
            self._write_lock = NoopLock()

    def emit(self, record):
        with self._write_lock:
            super().emit(record)

# A dummy "Lock" that does nothing when it is acquired and released using "with"
class NoopLock:
    def __enter__(self, *args, **kwargs): pass
    def __exit__(self, *args, **kwargs): pass



from silence.sql.tables import get_tables, get_views, get_primary_key, is_auto_increment, get_primary_key_views
from silence.logging.default_logger import logger
from silence.settings import settings

from os import listdir, getcwd, path, mkdir, makedirs
from shutil import rmtree
import json
import re

# Entry point for the CLI command
def create_api():
    existing_routes_method_pairs = get_user_endpoints()
    logger.info("Found the following user defined endpoints:")
    for rmp in existing_routes_method_pairs:
        logger.info(rmp)

    create_entity_endpoints(existing_routes_method_pairs)

#############################################################################
# Get the entities from the database and the existing user endpoints and    #
# create CRUD endpoint files (json) for the remaining ones.                 #
#############################################################################
def create_entity_endpoints(existing_routes_method_pairs):
    '''
    Reset the "auto" folder for endpoints and iterate over the databases tables and views,
    generating the corresponding endpoints for each of them.
    '''
    # Remove the folder if it exists and create it again
    auto_dir = getcwd() + "/endpoints/auto"
    
    try:
        rmtree(auto_dir)
    except FileNotFoundError:
        logger.debug("Endpoint folder not found.")
    
    logger.debug("Creating directory %s", auto_dir)
    makedirs(auto_dir)

    # Create the endpoints (JSON and JS) for the tables and views
    # We combine the tables and views in one list of tuples,
    # adding a "is_table" flag to distinguish them.
    tables_data = [(td, True) for td in get_tables().items()]
    views_data = [(vd, False) for vd in get_views().items()]
    all_data = tables_data + views_data

    for (table_name, columns), is_table in all_data:
        # We always write all autogenerated endpoints to the JS files,
        # while only new endpoints that have not been defined by the
        # user are added as JSON config files for the server
        endpoints_to_js = {}
        endpoints_to_json = {}
        logger.info("Generating endpoints and JS API files for %s", table_name)

        if is_table:
            pk = get_primary_key(table_name)
            route_one = f"/{table_name.lower()}/${pk}"

            if pk and is_auto_increment(table_name, pk):
                columns.remove(pk)

        # Create the basic CRUD endpoints if they haven't been defined yet
        route_all = f"/{table_name.lower()}"
        crud_endpoints = [
            ("GET", "getAll", route_all),
            ("GET", "getById", route_one),
            ("POST", "create", route_all),
            ("PUT", "update", route_one),
            ("DELETE", "delete", route_one),
        ]

        for method, name, route in crud_endpoints:
            # If this is a view or a table without a PK, create only "getAll"
            if name != "getAll" and (not is_table or pk is None): 
                continue

            endpoints_to_js[name] = make_endpoint_data(name, route, method, table_name, columns, pk)
            if (route, method) not in existing_routes_method_pairs:
                endpoints_to_json[name] = endpoints_to_js[name]

        # If this is a view, we generate additional JS functions based on the
        # primary keys from other tables it contains, for easier querying
        if not is_table:
            for pk in get_primary_key_views(table_name):
                endpoint_name = "getBy" + pk[0].upper() + pk[1:]  # Is there a better way to do this?
                endpoints_to_js[endpoint_name] = {
                    "route": f"{route_all}?{pk}=${pk}",
                    "method": "GET",
                    "description": f"Gets an array entries from '{table_name}' by their {pk}.\n{' '*8}Note that this always returns an array."
                }

        # Create *all* the .js files for the API.
        generate_API_file_for_endpoints(endpoints_to_js, table_name, pk)

        # Save the *new* endpoints as JSON config to a file
        if endpoints_to_json:
            json_str = json.dumps(endpoints_to_json, indent=4)
            with open(auto_dir + f"/{table_name}.json", "w", encoding="utf8") as f:
                f.write(json_str)

    # Finally, generate the .js API module for the allowed auth operations
    auth_endpoints = {}
    if settings.ENABLE_LOGIN:
        auth_endpoints["login"] = { "route": "/login", "method": "POST", "description": "Logs in using an identifier and password" }

    if settings.ENABLE_REGISTER:
        auth_endpoints["register"] = { "route": "/register", "method": "POST", "description": "Registers a new user and stores the password safely in the database" }

    if auth_endpoints:
        generate_API_file_for_endpoints(auth_endpoints, "auth", None)

###############################################################################
# Creates a dict with the auto-generated data for a basic CRUD endpoint
##############################################################################
def make_endpoint_data(name, route, method, table_name, cols, pk):
    sql, desc = {
        "getAll": (
            f"SELECT * FROM {table_name}",
            f"Gets all entries from '{table_name}'"
        ),
        "getById": (
            f"SELECT * FROM {table_name} WHERE {pk} = ${pk}",
            f"Gets an entry from '{table_name}' by its primary key",
        ),
        "create": (
            f"INSERT INTO {table_name} " + params_to_string(cols) + " VALUES " + params_to_string(cols, "$"),
            f"Creates a new entry in '{table_name}'"
        ),
        "update": (
            f"UPDATE {table_name} SET " + params_to_string(cols, is_update=True) + f" WHERE {pk} = ${pk}",
            f"Updates an existing entry in '{table_name}' by its primary key"
        ),
        "delete": (
            f"DELETE FROM {table_name} WHERE {pk} = ${pk}",
            f"Deletes an existing entry in '{table_name}' by its primary key"
        ),
    }[name]

    data = {
        "description": desc,
        "route": route,
        "method": method,
        "sql": sql,
        "request_body_params": cols,
        "auth_required": method != "GET",
        "allowed_roles": ["*"],
    }

    if method not in ("POST", "PUT"):
        del data["request_body_params"]
    
    return data

###############################################################################
# Create generic .js files to consume the created endpoints.
###############################################################################  
def generate_API_file_for_endpoints(endpoints, table_name, pk_name):
    api_path = getcwd() + "/web/js/api"
    
    if not path.isdir(api_path):
        makedirs(api_path)

    functions = [generate_api_text(ep_name, pk_name, ep_data) for ep_name, ep_data in endpoints.items()]
    functions_str = "\n\n    ".join(functions)

    file_content = f"""/*
 * DO NOT EDIT THIS FILE, it is auto-generated. It will be updated automatically.
 * All changes done to this file will be lost upon re-running the 'silence createapi' command.
 * If you want to create new API methods, define them in a new file.
 *
 * Silence is built and maintained by the DEAL research group at the University of Seville.
 * You can find us at https://deal.us.es
 */

"use strict";

import {{ BASE_URL, requestOptions }} from './common.js';

const {table_name}API_auto = {{

    {functions_str}
}};

export {{ {table_name}API_auto }};"""

    open(f"{api_path}/_{table_name}.js", "w", encoding="utf8").write(file_content)


def generate_api_text(name, pk_name, endpoint_data):
    method = endpoint_data["method"].lower()
    route = endpoint_data["route"]
    description = endpoint_data["description"]

    # Replace the URL params with JS's string interpolation method
    if "$" in route:
        route = re.sub(r"\$(\w+)", r"${\1}", route)

    is_getById = name == "getById"
    needs_formdata = method in ("post", "put")
    formdata = ", formData" if needs_formdata else ""

    args = []
    if needs_formdata: 
        args.append("formData")

    if method in ("put", "delete") or is_getById: 
        args.append(pk_name)
    
    return f"""/** {description} */
    {name}: async function({', '.join(args)}) {{
        let response = await axios.{method}(`${{BASE_URL}}{route}`{formdata}, requestOptions);
        return response.data{"[0]" if is_getById else ""};
    }},"""

def get_user_endpoints():
    '''
    Searches in the endpoints directory of the silence project for every .json file which we assume is an endpoint file.
    gets the route and method as pairs for every empoint found and returns them
    '''
    logger.debug("Looking for user endpoints")
    endpoints_dir = getcwd() + "/endpoints"

    # Create the endpoints dir if its not there
    if not path.isdir(endpoints_dir):
        mkdir(endpoints_dir)

    # Get every .json file in the directory
    user_endpoints = [endpoints_dir + f"/{f}" for f in listdir(endpoints_dir) if f.endswith(".json")]
    endpoint_route_method_pairs = []

    # Form (route, method) pairs for every endpoint found and return them
    for jsonfile in user_endpoints:
        with open(jsonfile, "r", encoding="utf8") as ep:
            endpoints = list(json.load(ep).values())
            endpoint_pair = [(endpoint['route'], endpoint['method']) for endpoint in endpoints]
            endpoint_route_method_pairs += endpoint_pair

    return endpoint_route_method_pairs

def params_to_string(param_list, char_add="", is_update=False):
    add_pref = lambda x: char_add + x if not is_update else f"{x} = ${x}"
    res = ", ".join(add_pref(x) for x in param_list)
    return f"({res})" if not is_update else res


from silence.settings import settings

import colorama
from colorama import Fore, Style

from datetime import datetime
import logging
import re

###############################################################################
# Filters and modifies Flask's log records in-place
###############################################################################

colorama.init()

# Regex to remove ANSI color codes from log lines
RE_ANSI = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')

# Regex to detect server log lines from Flask (werkzeug)
RE_LOG = re.compile(r'(.*) - - \[.*\] "%s" %s %s')
RE_REQ_DATA = re.compile(r'(\w+) (/.*) HTTP.*')

COLORS = {
    "GREEN": Style.BRIGHT + Fore.GREEN if settings.COLORED_OUTPUT else "",
    "MAGENTA": Style.BRIGHT + Fore.MAGENTA if settings.COLORED_OUTPUT else "",
    "CYAN": Style.BRIGHT + Fore.CYAN if settings.COLORED_OUTPUT else "",
    "YELLOW": Style.BRIGHT + Fore.YELLOW if settings.COLORED_OUTPUT else "",
    "RED": Style.BRIGHT + Fore.RED if settings.COLORED_OUTPUT else "",
    "WHITE": Style.BRIGHT + Fore.WHITE if settings.COLORED_OUTPUT else "",
}
RESET = Style.RESET_ALL if settings.COLORED_OUTPUT else ""

class FlaskFilter(logging.Filter):

    def filter(self, record):
        msg = record.msg

        if msg.startswith(" * Running on"):
            record.msg = msg[3:]
            return True

        # If the log line is a Flask log line, parse it and
        # style it appropriately
        m = RE_LOG.match(msg)
        if m:
            # The address is parsed from the message
            addr = m.group(1)

            # The rest of the data comes from the args tuple in the log record
            args = record.args
            code = args[1]

            # The address and the route are present in the first arg,
            # but we must remove color codes first before parsing
            addr_and_route = RE_ANSI.sub("", args[0])
            m_addr = RE_REQ_DATA.match(addr_and_route)
            verb = m_addr.group(1)
            route = m_addr.group(2)

            if route.startswith(settings.API_PREFIX):
                api_web = "[API]"
                api_color = COLORS["MAGENTA"]
            else:
                api_web = "[WEB]"
                api_color = COLORS["CYAN"]

            if code[0] in ('2', '3'):
                code_color = COLORS["GREEN"]
            elif code[0] == '4':
                code_color = COLORS["YELLOW"]
            elif code[0] == '5':
                code_color = COLORS["RED"]
            else:
                code_color = COLORS["WHITE"]

            # Finally, the date is created from scratch here
            date = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

            record.msg = f"{date} | {api_color}{api_web}{RESET} " + \
                  f"{verb} {route} from {addr} - {code_color}{code}{RESET}"
            record.args = ()

        # Pass it on to Silence's logger, which implements
        # blocking to prevent multiple lines from overlapping,
        # and filter it out from this one
        logging.getLogger("silence").handle(record)
        return False



import sys
import re

from flask import jsonify, request

from silence.db                         import dal
from silence                            import sql as SQL
from silence.sql                        import get_sql_op
from silence.settings                   import settings
from silence.utils.min_type             import Min
from silence.auth.tokens                import check_token
from silence.sql.tables                 import get_primary_key
from silence.logging.default_logger     import logger
from silence.sql.converter              import silence_to_mysql
from silence.server                     import manager as server_manager
from silence.exceptions                 import HTTPError, TokenError

OP_VERBS = {
    SQL.SELECT: 'get',
    SQL.INSERT: 'post',
    SQL.UPDATE: 'put',
    SQL.DELETE: 'delete',
}

RE_QUERY_PARAM = re.compile(r"^.*\$\w+/?$")

###############################################################################
# This is where the fun at v2
###############################################################################

def setup_endpoint(route, method, sql, auth_required=False, allowed_roles=["*"], description=None, request_body_params = []):
    logger.debug("Setting up endpoint %s %s", method, route)
    
    # Check if the query is requesting the logged user's ID
    # If it is and the endpoint does not required authentication,
    # warn that it may be NULL
    logged_user = "$loggedId" in sql
    if logged_user and not auth_required:
        logger.warning("The endpoint %s %s uses $loggedId but does not require authentication.\n Keep in mind that the logged user's ID may be NULL.", method.upper(), route)

    # Construct the API route taking the prefix into account
    route_prefix = settings.API_PREFIX
    if route_prefix.endswith("/"):
        route_prefix = route_prefix[:-1]  # Drop the final /
    full_route = route_prefix + route

    # Warn if the pair SQL operation - HTTP verb is not the proper one
    check_method(sql, method, route)

    # Warn if the values of auth_required and allowed_roles don't make sense together
    check_auth_roles(auth_required, allowed_roles, method, route)

    # Extract the list of parameters that the user expects to receive
    # in the URL and in the SQL string
    sql_params = extract_params(sql)
    url_params = extract_params(route)

    # Get the required SQL operation
    sql_op = get_sql_op(sql)

    # If it's a SELECT or a DELETE, make sure that all SQL params can be
    # obtained from the url
    if sql_op in (SQL.SELECT, SQL.DELETE):
        check_params_match(sql_params, url_params, route)

    # If it's a SELECT or a DELETE, make sure that all SQL params can be
    # obtained from the url AND the request body
    if sql_op in (SQL.INSERT, SQL.UPDATE):
        check_params_match(sql_params, url_params + request_body_params, route)

    # The handler function that will be passed to flask
    def route_handler(*args, **kwargs):
        # Get the user's data from the session token
        logged_user_data = get_logged_user()

        # If this endpoint requires authentication, check that the
        # user is logged in and has the proper role
        if auth_required:
            check_session(logged_user_data, allowed_roles)
        
        # Collect all url pattern params
        request_url_params_dict = kwargs

        # If the endpoint requires the logged user's ID, we add it to the available params
        if logged_user:
            request_url_params_dict["loggedId"] = get_current_user_id(logged_user_data)

        # Convert the Silence-style placeholders in the SQL query to proper MySQL placeholders
        query_string = silence_to_mysql(sql)

        # Default outputs
        res = None
        status = 200
        
        # SELECT/GET operations
        if sql_op == SQL.SELECT:
            # The URL params have been checked to be enough to fill all SQL params
            url_pattern_params = tuple(request_url_params_dict[param] for param in sql_params)
            res = dal.api_safe_query(query_string, url_pattern_params)
            
            # Filter these results according to the URL query string, if there is one
            # Possible TO-DO: do this by directly editing the SQL query for extra efficiency
            res = filter_query_results(res, request.args)

            # In our teaching context, it is safe to assume that if the URL ends
            # with a parameter and we have no results, we should return a 404 code
            if RE_QUERY_PARAM.match(route) and not res:
                raise HTTPError(404, "Not found")

        else:  # POST/PUT/DELETE operations
            #Construct a dict for all params expected in the request body, setting them to None if they have not been provided
            form = request.json if request.is_json else request.form
            body_params = {param: form.get(param, None) for param in request_body_params}

            # We have checked that sql_params is a subset of url_params U body_params,
            # construct a joint param object and use it to fill the SQL placeholders
            for param in url_params:
                body_params[param] = request_url_params_dict[param]

            param_tuple = tuple(body_params[param] for param in sql_params)

            # Run the execute query
            res = dal.api_safe_update(query_string, param_tuple)

        return jsonify(res), status
    
    # flaskify_url() adapts the URL so that all $variables are converted to Flask-style <variables>
    server_manager.APP.add_url_rule(flaskify_url(full_route), method + route, route_handler, methods=[method])
    server_manager.API_SUMMARY.register_endpoint({"route": full_route, "method": method.upper(), "description": description})

###############################################################################
# Aux functions

# Extracts the user's data from the session token sent in the header
# Returns a dict with the user's information, or None if the token
# is not present or is invalid
def get_logged_user():
    token = request.headers.get("Token", default=None)
    logged_user_data = None

    if token:
        try:
            logged_user_data = check_token(token)
        except TokenError as exc:
            logger.debug("The user sent an invalid token: %s", str(exc))

    return logged_user_data

# Checks whether the user is logged in and has the proper role
# Raises a 401 HTTP error if the previous conditions are not met
# Note that user_data may be null if the user has not sent a token
# or has sent an invalid one
def check_session(logged_user_data, allowed_roles):
    if logged_user_data is None:
        raise HTTPError(401, "Unauthorized")

    # Check if the user's role is allowed to access this endpoint
    role_col_name = settings.USER_AUTH_DATA.get("role", None)
    
    if role_col_name: # Only check the role if we know the role column
        # Find the role of the user from the user data
        user_role = next((v for k, v in logged_user_data.items() if k.lower() == role_col_name.lower()), None)

        logger.debug("Allowed roles are %s and the user role is %s", str(allowed_roles), user_role)

        if user_role not in allowed_roles and "*" not in allowed_roles:
            raise HTTPError(401, "Unauthorized")
    
# Return the user's ID if the user is logged in and has a PK, None otherwise
def get_current_user_id(logged_user_data):
    userID = None

    if logged_user_data is not None:
        users_table = settings.USER_AUTH_DATA['table']
        pk = get_primary_key(users_table)
        userID = logged_user_data[pk] if pk else None  # Returns None if the table has no primary key

    return userID

# Implements filtering, ordering and paging using query strings
def filter_query_results(data, args):
    # Grab all parameters from the query string
    sort_param = args.get("_sort")
    sort_reverse = args.get("_order") == "desc"

    try:
        limit = int(args.get("_limit", None))
    except (ValueError, TypeError):
        limit = None

    try:
        page = int(args.get("_page", None))
    except (ValueError, TypeError):
        page = None

    # Filter, sort and paginate results
    filter_criteria = [pair for pair in args.items() if not pair[0].startswith("_")]
    filter_func = lambda elem: all(k not in elem or str(elem[k]).lower() == v.lower() for k, v in filter_criteria)
    res = list(filter(filter_func, data))

    def order_key_func(elem):
        v = elem[sort_param]
        return v if v is not None else Min  # Avoids errors when comparing against None

    try:
        res.sort(key=order_key_func, reverse=sort_reverse)
    except KeyError: pass

    offset = limit * page if limit and page else 0
    top = offset + limit if limit else len(res)
    return res[offset:top]

# Checks whether the SQL operation and the HTTP verb match
def check_method(sql, verb, endpoint):
    sql_op = get_sql_op(sql)

    if sql_op in OP_VERBS:
        correct_verb = OP_VERBS[sql_op]

        if correct_verb != verb.lower():
            # Warn the user about the correct verb to use
            logger.warn(
                f"The '{verb.upper()}' HTTP verb is not correct for the SQL {sql_op.upper()} " +
                f"operation in endpoint {verb.upper()} {endpoint}, the correct verb is {correct_verb.upper()}."
            )
    else:
        # What has the user put here?
        logger.error(f"The SQL query '{sql}' in the endpoint {endpoint} is not supported," +
                             " please use only SELECT/INSERT/UPDATE/DELETE.")
        sys.exit(1)

def check_auth_roles(auth_required, allowed_roles, method, route):
    # Raise an error if allowed_roles is not a list
    if not isinstance(allowed_roles, list):
        logger.error(f"The value '{allowed_roles}' for the allowed_roles parameter in " +
                            f"endpoint {method.upper()} {route} is not allowed, it must be a " +
                             "list of allowed roles.")
        sys.exit(1)

    # Warn if the user has specified some roles but auth_required is false,
    # since it will result in all users having access
    if not auth_required and len(allowed_roles) > 0 and allowed_roles != ["*"]:
        logger.warn(f"You have specified allowed roles in endpoint {method.upper()} {route}, " +
                     "but auth_required is False. This will result in all users having access " +
                     "regardless of their role.")

    # Warn if the user has specified an empty list of roles, and auth_required is true,
    # because it will result in noone having access
    if auth_required and len(allowed_roles) == 0:
        logger.warn(f"You have set auth_required to True in endpoint {method.upper()} {route}, " +
                     "but the list of allowed roles is empty. This will result in noone being able " +
                     "to access the endpoint.")

# Returns a list of $params in a SQL query or endpoint route,
# without the $'s
def extract_params(string):
    res = re.findall(r"\$\w+", string)
    return [x[1:] for x in res]

# Convers $url_param to <url_param>
def flaskify_url(url):
    return re.sub(r"\$(\w+)", r"<\1>", url)

# Checks whether all SQL params can be filled with the provided params
def check_params_match(sql_params, user_params, route):
    sql_params_set = set(sql_params)
    user_params_set = set(user_params)
    diff = sql_params_set.difference(user_params_set)
    if "loggedId" in diff:
        diff.remove("loggedId")

    if diff:
        params_str = ", ".join(f"${param}" for param in diff)
        logger.error(f"Error creating endpoint {route}: the parameters " +
                            f"{params_str} are expected by the SQL query but they are not provided in the URL " +
                            "or the request body.")
        sys.exit(1)


import importlib
import pprint
import sys
import os

from silence.settings import default_settings

###############################################################################
# A Settings class that contains the default settings as attributes,
# overriding them with the user provided settings when necessary.
###############################################################################

# If you are editing this file, keep this in mind:
# This is imported by many different parts of Silence to read the settings
# of the current project. This means that importing other parts of Silence
# here will most likely result in an error caused by a circular import.
# That's the reason why we can't even import the logger here and we resort
# to print(), because the logger itself needs the settings to be fully loaded
# to configure itself.
# So, try to keep the imports of other Silence modules to a bare minimum in
# the headers of this file. Importing them inside functions is OK, provided
# that they will be called after everything has been loaded

class Settings:
    def __init__(self):
        self.setting_keys = set()
        self.unknown_settings = set()

        # Load the default settings
        for setting in dir(default_settings):
            if setting.isupper():
                self.setting_keys.add(setting)
                setattr(self, setting, getattr(default_settings, setting))

        # Override the default settings with the user-defined ones
        try:
            sys.path.append(os.getcwd())
            user_settings = importlib.import_module("settings")

            for setting in dir(user_settings):
                # If the name of the setting is not known, store it to
                # later warn the user about them (unless it's the SECRET_KEY,
                # which has no default)
                if setting != "SECRET_KEY" and not hasattr(default_settings, setting):
                    self.unknown_settings.add(setting)

                if setting.isupper():
                    self.setting_keys.add(setting)
                    setattr(self, setting, getattr(user_settings, setting))

            # Ensure that the user has provided a SECRET_KEY
            if not hasattr(self, "SECRET_KEY"):
                print("You need to provide a SECRET_KEY in your settings.py file.")
                print("If you have git clone'd this project, use 'silence new' instead,")
                print("and it will automatically generate a SECRET_KEY for you.")
                sys.exit(1)

        except ModuleNotFoundError:
            if "run" in sys.argv or "createdb" in sys.argv:
                print("Can't find the settings.py file! Are you running this command from the project's root folder?")
                sys.exit(1)

    def warn_unknown_settings(self):
        from silence.logging.default_logger import logger
        if self.unknown_settings:
            logger.warning("The following parameters in your settings.py are " 
                           "unknown and will be ignored: " + ", ".join(self.unknown_settings))

    def __str__(self):
        return pprint.pformat({setting: getattr(self, setting) for setting in self.setting_keys})
        
settings = Settings()



# Thanks to https://stackoverflow.com/a/3885198/5604339 :)
import decimal
from flask import json

from silence.settings import settings

class SilenceJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            func = str if settings.DECIMALS_AS_STRINGS else float
            return func(o)
        elif hasattr(o, "isoformat"):
            # Dates, times and datetimes should be formatted following ISO-8601
            # with the format YYYY-MM-DD HH:MM:SS (or only the corresponding fragment
            # for dates and times)
            # ISO-8601 also defines "T" as a separator between date and time, which
            # we replace by a space for compatibility with MySQL
            # Instead of checking directly for their types, we test whether they
            # have the isoformat() method, which all date/time classes share
            return o.isoformat().replace("T", " ")

        try:
            return super().default(o)
        except TypeError:
            return str(o)

class SilenceJSONSerializer:
    def dumps(o):
        return json.dumps(o, cls=SilenceJSONEncoder)

    def loads(s):
        return json.loads(s)


from silence.logging.default_logger import logger
from silence.settings import settings
from silence.utils.check_update import check_for_new_version
from silence import __version__


def handle(args):
    from silence.server import manager as server_manager
    
    logger.info("Silence v%s", __version__)
    logger.debug("Current settings:\n%s", str(settings))

    new_ver = check_for_new_version()
    if new_ver:
        logger.warning("A new Silence version (v%s) is available. Run 'pip install --upgrade Silence' to upgrade.", new_ver)
        logger.warning("To see what's new, visit: https://github.com/DEAL-US/Silence/blob/master/CHANGELOG.md")

    server_manager.setup()
    server_manager.run()


from os import getcwd
from silence.sql.tables import get_tables, get_primary_key, get_table_cols
from silence.logging.default_logger import logger
from silence.settings import settings
from shutil import rmtree
from pathlib import Path

# SILENCE CREATETEST OPERATIONS

# Entry point for CLI command
###########################################
# Get the entities from the database and  #
# create CRUD test files (http) for them. #
###########################################
def create_tests():
    # Folder handling
    curr_dir = getcwd()
    test_dir = curr_dir + "/tests/auto"

    logger.debug("Selected test directory: %s", test_dir)
    try:
        rmtree(test_dir)
    except FileNotFoundError:
        logger.debug("Folder is not there, creating it.")

    logger.debug("Re-creating directory %s", test_dir)
    Path(test_dir).mkdir(parents=True, exist_ok=True)

    # Test files creation
    tables = get_tables()
    table_name_auth = settings.USER_AUTH_DATA["table"].lower()

    for table in list(tables.items()):
        name = table[0].lower()
        pk = get_primary_key(name)
        if pk is None:
                logger.warning("The table '%s' does not have a primary key. No tests will be created for it.", name)
                continue
        try:
            table[1].remove(pk) # The auth table will already have its primary key removed.
        except:
            pass
        
        logger.info("Generating tests for %s", name)
        table_name = next(t for t in tables if t.lower() == name)
        table_attributes = get_table_cols(table_name)

        TEST = f"""
### This is an auto-generated test suite, it needs to be completed with valid data.
### These are not all tests you need, more of them should be created to evaluate the functional
### requirements of your project. These tests only test the CRUD endpoints of the entity.
### Silence is a DEAL research team project, more info about us in https://deal.us.es
@BASE = http://{settings.LISTEN_ADDRESS}:{settings.HTTP_PORT}{settings.API_PREFIX}

### Auxiliary query
### Positive test
### Test 00: Get all existing {name}
### This query is used in several of the below tests it should be executed first.
# @name {name}
GET {{{{BASE}}}}/{name}

### Login a(n) {table_name_auth} and save the generated token 
### This token is used in several of the below tests it should be executed second.
# @name login
POST {{{{BASE}}}}/login 
Content-Type: application/json 

{{ 
"""
        t_args = [settings.USER_AUTH_DATA["identifier"], settings.USER_AUTH_DATA["password"]]
        TEST += add_table_args(t_args)
        TEST += f"""}}

###
@token = {{{{login.response.body.sessionToken}}}}
"""
        TEST += f"""

### TESTS BEGIN

### Test 01: Get one existing {name} by its id.
### positive test 
@{name[:3]}Id = {{{{{name}.response.body.0.{pk}}}}}
GET {{{{BASE}}}}/{name}/{{{{{name[:3]}Id}}}} 
Content-Type: application/json 


### Test 02: Try get one existing {name} by its nonexistent id.
### negative test 
GET {{{{BASE}}}}/{name}/999999999
Content-Type: application/json 


### Test 03: Add a new {name} successfully
### Positive test 
### We assume that the token has been aquired by the login request.
# @name new{name}
POST {{{{BASE}}}}/{name} 
Content-Type: application/json 
Token: {{{{token}}}} 

{{
"""
        TEST += add_table_args(table_attributes)
        TEST += f"""}} 

### Check the created {name}

@new{name}id = {{{{new{name}.response.body.lastId}}}}
GET {{{{BASE}}}}/{name}/{{{{new{name}id}}}}
Content-Type: application/json 

### Test 04: Add a new {name} without a session token
### Negative test 
POST {{{{BASE}}}}/{name} 
Content-Type: application/json 

{{ 
"""
        TEST += add_table_args(table_attributes)
        TEST += f"""}}

### Test 05: Modify an existing {name} 
### Positive test 
@{name[:3]}Id = {{{{{name}.response.body.0.{pk}}}}} 
PUT {{{{BASE}}}}/{name}/{{{{{name[:3]}Id}}}} 
Content-Type: application/json 
Token: {{{{token}}}} 

{{ 
"""

        TEST += add_table_args(table_attributes)
        TEST += f"""}}

### Check the modified {name}

GET {{{{BASE}}}}/{name}/{{{{new{name}id}}}}
Content-Type: application/json 

### Test 06: Try to modify an existing {name} without a session token
### Negative test
@{name[:3]}Id = {{{{{name}.response.body.0.{pk}}}}}
PUT {{{{BASE}}}}/{name}/{{{{{name[:3]}Id}}}}
Content-Type: application/json

{{
"""
        TEST += add_table_args(table_attributes)
        TEST += f"""}}

### Test 07: Delete an existing {name}
### Positive test

### Create a new {name}, which we will delete
# @name created{name[:3]}ToDelete
POST {{{{BASE}}}}/{name}
Content-Type: application/json
Token: {{{{token}}}}

{{
"""
        TEST += add_table_args(table_attributes)
        TEST += f"""}}

### Check the created {name}
@{name[:3]}delId = {{{{created{name[:3]}ToDelete.response.body.lastId}}}}
GET {{{{BASE}}}}/{name}/{{{{{name[:3]}delId}}}}
Content-Type: application/json 

### Delete the {name}
DELETE {{{{BASE}}}}/{name}/{{{{{name[:3]}delId}}}}
Token: {{{{token}}}}

### Check the deleted {name}
GET {{{{BASE}}}}/{name}/{{{{{name[:3]}delId}}}}
Content-Type: application/json 

### Test 08: Try to delete a {name} without a session token
### Negative test
DELETE {{{{BASE}}}}/{name}/{{{{{name[:3]}Id}}}}

"""
        # Write test to file
        with open(test_dir + f"/{name}.http", "w", encoding="utf8") as test:
            test.write(TEST)

def add_table_args(table_attributes):
    res = ""
    for i, t in enumerate(table_attributes):
        if i != len(table_attributes) - 1:
            res += f"\t\"{t}\": ###REPLACE###,\n"
        else:
            res += f"\t\"{t}\": ###REPLACE###\n"
    return res